/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (input.C)
  
  Written by: Damian J. J. Farnell

  Date:       Wed Jul 17 12:03:05 BST 2002

  Purpose:    Inputs the CCM script data

  Changes:
      - cout replaced by printf, little improvements (Joerg Sep2005)

  ===================================================

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

FILE *script;
FILE *escript;

// function prototypes.

int   moveAlongHome(FILE *fp, const char *searchstring);
double  readInFloat(FILE *fp);
int   readInInteger(FILE *fp);

void   definePointsInBT(void);

void input(int run_option, char *ccmfile) 
{
  
  int i, j ,k, bonds, expect, max_s2;
  int ierr=0; // error value for last moveAlongHome()
  
  // initialized with "" to avoid warnings from valgrind
  //     "jump depend from uninitialized variable"
  char input_filename[100]="";  // $(arg1) = ./workdir/name.ccm
  // char tempPath[100]="";    // old options, obsolete
  // char temp[100]="";          
  char expectations_input_filename[100]=""; // $(arg1)_exp
  char configurations2[100]="";

  float dummyFloat;

  Control.file=1;    // tells program to save CCM equations to file.

  if (myid==0)
  printf("/-------------------------------------------------\\\n"
	 "|                                                 |\n"
	 "|                                                 |\n"
	 "|     /=====\\     /=====\\     |\\          /|      |\n"
	 "|    /           /            | \\        / |      |\n"
	 "|    |           |            |  \\      /  |      |\n"
	 "|    |           |            |   \\    /   |      |\n"
	 "|    \\           \\            |    \\  /    |      |\n"
	 "|     \\=====/     \\=====/     |     \\/     |      |\n"
	 "|                                                 |\n"
	 "|                                                 |\n"
	 "|           =====>>>>> 2002 <<<<=====             |\n"
	 "|                                                 |\n"
	 "\\-------------------------------------------------/\n"
	 "\n"
	 "===================================================\n"
	 "Name:       ccm.C\n"
	 "@Copyright: Damian J. J. Farnell\n"
	 "Date:       Wed Jul 17 12:03:05 BST 2002\n"
	 "Purpose:    Determine and solve CCM LSUBm equations.\n"
	 "===================================================\n\n");

  if (myid==0) {
    printf(" name=%s runoption=%d\n", ccmfile, run_option);
    printf("\n  v0806: all other file names are build by $(name) + suffix\n");
    printf("  this is incompatible to older versions but easier to handle\n");
    // especially for jobsystems
  }
  sprintf(input_filename,              "%s",    ccmfile);
  sprintf(expectations_input_filename, "%s_exp", ccmfile);

  script=fopen(input_filename,"r");
  if (!script) {
    if (!myid) {
      printf("input: Cannot Open %20s ...\n", input_filename);
      if (strchr(ccmfile,'/')==NULL)
        printf("v0806: Please include the path name to the filename\n");
    }
    ERR(1);
  }

  // read in `control' parameters first:

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "Control Parameters:\n");

  moveAlongHome(script,"output level"); Control.output=readInInteger(script);

  if (Control.output > 0 && myid ==0)
    printf("Output (0-non to >6 verbose) =%d\n",Control.output);

  // new by Jun08:
  ierr=moveAlongHome(script,"find generalised expectation values");
  if (!ierr) Control.greatExpectations=readInInteger(script);
  else       Control.greatExpectations=0;

  // Joerg Jun08: better get it as an option than changing ccm format again and again
  // moveAlongHome(); Control.excitations=readInInteger();

  if (Control.output > 0 && myid ==0)
    printf("Find Excitations (0-No : 1-Yes) =%d (Runoption = x)\n",
           Control.excitations);

  if (Control.output > 0 && myid ==0)
    printf("Find Generalised Expectations (0-No : 1-Yes) =%d\n",
                Control.greatExpectations);

  if (myid==0 && Control.output > 0 && Control.greatExpectations > 0)
    printf("myid=%d expectations name=%s\n", myid, expectations_input_filename);

  if (Control.greatExpectations > 0) {
    escript = fopen(expectations_input_filename,"r");
    if (!escript) {  // if that happens on some tasks we are in trouble
      if (myid<4)
        printf("WARNING: Cannot Open %20s ..."
               " disable computation of expectations\n",
                 expectations_input_filename);
      // disable or exit?
      Control.greatExpectations = 0;
    }
  }

  if (!ierr)  // skip if the "expectation" options was missed (old versions)
  moveAlongHome(script,"work directory"); // fscanf(script,"%80s",tempPath);  // work directory= ...
  moveAlongHome(script,"output file");    // fscanf(script,"%80s",temp);      // output file= ...
  // this options will be ignored, tempPath is unused
  
  // sprintf(Control.name,"%s%s_id%03d", tempPath, temp, myid);
  // sprintf(Control.name, "%s_id%03d", ccmfile, myid); // "name.ccm_id$(id)"
  // JS0902: only use _MYID for splitted files (.terms_MYID)
  sprintf(Control.name, "%s", ccmfile); // JS0902: replace "name.ccm_id$(id)"

  if (myid==0)
    printf("Simulation Filename = %s\n"
           " v0806: path and file name options from input file are ignored\n",
            Control.name);

  // read in parameters for solve.C
  
  // now read in the `solve' variables

  moveAlongHome(script,"minimum");   minimum=readInFloat(script);
  moveAlongHome(script,"maximum");   maximum=readInFloat(script);
  moveAlongHome(script,"increment"); increment=readInFloat(script);
  if (Control.output > 0 && myid ==0) {
    printf("Minimum delta=  %20.10f\n",minimum);
    printf("Maximum delta=  %20.10f\n",maximum);
    printf("Increment delta=%30.20f\n",increment);
  }

  // next two variables are minimum and maximum y values
  // used in plot.tcl but not in ccm C program.

  moveAlongHome(script,"minimumy"); dummyFloat=readInFloat(script);
  moveAlongHome(script,"maximumy"); dummyFloat=readInFloat(script);

  // read in highest level parameters first:

  if (Control.output > 0 && myid ==0)
    printf("\nGeneral Parameters:\n");
  
  // dimension

  moveAlongHome(script,"dimension"); Params.dim=readInInteger(script);
  if (Control.output > 0 && myid ==0)
    printf("Dimension=%d\n", Params.dim);

  if (Params.dim != Dimension) {
    printf("Error: Constant.h Dimension need to be set to %d for this model\n"
           "  higher values are not allowed, we waste energy + time else\n",
           Params.dim);
    if (Params.dim > Dimension) MPI_Abort(MPI_COMM_WORLD, 1);
  }

  // spin number

  moveAlongHome(script,"spin"); dummyFloat = readInFloat(script);
  // Joerg: We store the twofold of Spin quantum number to have an integer
  //        this is faster and makes the program more compact
  // ToDo: its the maximum of spins in the unit cell and can be
  //       automatically calculated (see below), so its obsolete to set here
  Params.s2 = (int) (2*dummyFloat + 0.5);
  if (Control.output > 0 && myid ==0)
    printf("Maximum spin quantum number=%f\n",Params.s2 / 2.0);

  // Bounding Triangle

  // dummyInt=0;

  moveAlongHome(script,"bounding"); Params.btLength=readInInteger(script);
  if (Control.output > 0 && myid ==0)
    printf("BT Length=%d\n",Params.btLength);  // Joerg: bounding triangle
  // JS: length of triangle block >= m ? 
  
  // SUBn-m: n= spin flips
  // SUBn-m: m= max. geom. distance
  moveAlongHome(script,"n"); Params.n=readInInteger(script);
  moveAlongHome(script,"m"); Params.m=readInInteger(script);
  if (Control.output > -1 && myid ==0) // essential debug info
    printf("SUBn-m: n=%d m=%d\n", Params.n, Params.m);
  
  if (Params.btLength < Params.m) {
     if (!myid) printf("Warning input: bounding < m, set bounding = m\n");
     Params.btLength = Params.m; // Joerg Oct05, set bounding triangle to m
  }

  moveAlongHome(script,"bonds"); Params.bonds=readInInteger(script);
  if (Control.output > 0 && myid ==0)
    printf("Bonds: m=%d\n",Params.bonds);

  if (Params.bonds > MAX_BONDS) { // defined in Externals.h
    ERR("Error input: Bonds > MAX_BONDS\n");
  }

  moveAlongHome(script,"restriction"); restrictionClusters=readInInteger(script);
  if (Control.output > 0 && myid ==0)
    printf("Restriction on Clusters: Type=%d\n",restrictionClusters);
  
  moveAlongHome(script,"solution strategy"); solutionStrategy=readInInteger(script);
  if (PARALLEL_FLAG>0 && solutionStrategy == 1 && npe != 1) {
    if (myid==0)
      printf("WARNING: solutionStrategy 1 (Newton-Raphson) dont work parallel\n"
            "switching to solutionStrategy = 0 (Direct Iteration)\n");
    // old code exit here, but we make life easier (JS Apr08)
    solutionStrategy=0;
  }
  if (Control.output > 0 && myid ==0) {
    if (solutionStrategy == 1)
      printf("Solve CCM equations using Newton-Raphson method (1)\n");
    else
      printf("Solve CCM equations using Direct Iteration method (0)\n");
  }
  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Unit Cell:\n");

  // BT = bounding triangle  
  // number of different elements of a symmetric multidimensional matrix

  max_number_cells_BT = 1;  // js0903
  { int i;
    for (i=0; i<Params.dim; i++) max_number_cells_BT *= (Params.btLength+i);
    for (i=0; i<Params.dim; i++) max_number_cells_BT /= (i+1);
  }
  if (Control.output > -1 && myid ==0) 
    printf("\nDimension=%d  2s=%d  m=%d  btLength=sizeLUT=%d  "
	   "#Sites in BT.=%d\n\n",
	   Params.dim, Params.s2, Params.m, Params.btLength,
	   max_number_cells_BT);

  // probably a good idea to keep 
  // Params.btLength >= Params.m/UnitCell.number+1;

  for(i=0;i<MAX_UNIT_CELL;i++) {
    UnitCell.x[i]=0.0;
    UnitCell.y[i]=0.0;
    UnitCell.z[i]=0.0;
    UnitCell.s2[i]=1; // 2*s
  }

  moveAlongHome(script,"number"); fscanf(script,"%d",&UnitCell.number);
  if (Control.output > 0 && myid ==0)
    printf("Number of sites in Unit Cell = %d\n", UnitCell.number);
  if (Control.output > 0 && myid ==0)
    printf("Positions & Spin Quantum Number:\n");

  if (UnitCell.number>MAX_UNIT_CELL)
    { printf("Error: UnitCell.number>MAX_UNIT_CELL\nExit\n"); exit(1); }

  // read in all sites of the unit cell
  for(max_s2=1,i=1;i<=UnitCell.number;i++) {
    
    UnitCell.x[i]=readInFloat(script);
    UnitCell.y[i]=readInFloat(script);
    UnitCell.z[i]=readInFloat(script);
    UnitCell.s2[i]=(int)(2*fabs(readInFloat(script))+0.5);
    //    moveAlongHome(script);
    if (max_s2 < UnitCell.s2[i])
        max_s2 = UnitCell.s2[i];

  }

  // check that spin length doesnt exceed maxspinlength (Joerg 2005)
  if (max_s2 != Params.s2) {
     if (!myid) {
       printf("\nWARNING input: (max_spin=%3.1f) !=  (spin=%3.1f),"
              "\n set Params.s2 automatically to max_spin\n",
              0.5*max_s2, 0.5*Params.s2);
       printf("--- Do you forgot to set spin length in the unit cell? ---\n");
     }
     Params.s2 = max_s2;
     // exit(1); // We dont need to worry, Joerg
  }

  initDivisors(Params.s2);  // store 0! .. (2s+1)! into array

  for(i=1;i<=UnitCell.number;i++) {
    if (Control.output > 0 && myid ==0)
      printf("Site= %2d : position = (% f,% f,% f) : spin = %f\n",i,
	     UnitCell.x[i],
	     UnitCell.y[i],
	     UnitCell.z[i],
	     0.5*UnitCell.s2[i]);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Bravais Lattice:\n");

  // define the bravais lattice 

  moveAlongHome(script,"");
  for (i=0; i<3; i++) // assume 3D
  for (j=0; j<3; j++)
    BravaisLattice[i][j]=readInFloat(script); // old0903: .{a,b,c}[1..3]
  
  if (Control.output > -1 && myid ==0) {
    for (i=0; i<3; i++)
      printf("Bravais Lattice Vector %c=(% 6.4f,% 6.4f,% 6.4f)\n", 'A'+i,
  	      BravaisLattice[i][0],
              BravaisLattice[i][1],
              BravaisLattice[i][2]);
  }
  { // JS0903: compute inverse matrix to solve A.x = b later
    //  invers = adj(A)/det(A),  adj(A)=transpose(subDet)
    double det, non_ortho=0.0, (*m)[3]=BravaisLattice;
    for (i=0; i<3; i++)
      non_ortho += fabs(m[0][i]*m[1][i])
                 + fabs(m[1][i]*m[2][i])
                 + fabs(m[2][i]*m[0][i]);
    det =  m[0][0]*m[1][1]*m[2][2]
         + m[0][1]*m[1][2]*m[2][0]
         + m[0][2]*m[1][0]*m[2][1]
         - m[0][0]*m[1][2]*m[2][1]
         - m[0][2]*m[1][1]*m[2][0]
         - m[0][1]*m[1][0]*m[2][2];
    if (Control. output > -1 && myid ==0) 
      printf(" Bravais Lattice Orthogonality if 0 == %g\n"
             " Det([A,B,C])= %g\n", non_ortho, det);
     inversBravaisLattice[0][0] = (m[1][1]*m[2][2]-m[1][2]*m[2][1])/det;
     inversBravaisLattice[0][1] = (m[0][2]*m[2][1]-m[0][1]*m[2][2])/det;
     inversBravaisLattice[0][2] = (m[0][1]*m[1][2]-m[0][2]*m[1][1])/det;
     inversBravaisLattice[1][0] = (m[1][2]*m[2][0]-m[1][0]*m[2][2])/det;
     inversBravaisLattice[1][1] = (m[0][0]*m[2][2]-m[0][2]*m[2][0])/det;
     inversBravaisLattice[1][2] = (m[0][2]*m[1][0]-m[0][0]*m[1][2])/det;
     inversBravaisLattice[2][0] = (m[1][0]*m[2][1]-m[1][1]*m[2][0])/det;
     inversBravaisLattice[2][1] = (m[0][1]*m[2][0]-m[0][0]*m[2][1])/det;
     inversBravaisLattice[2][2] = (m[0][0]*m[1][1]-m[0][1]*m[1][0])/det;
  }
  if (Control.output > -1 && myid ==0) {
    for (i=0;i<3;i++)
      printf("invers Bravais Lattice Vector i%c=(% 6.4f,% 6.4f,% 6.4f)\n",
              'A'+i,
  	      inversBravaisLattice[i][0],
              inversBravaisLattice[i][1],
              inversBravaisLattice[i][2]);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "Nearest-Neighbours to sites in Unit Cell:\n");
  

  for(i=1;i<=UnitCell.number;i++) {
    moveAlongHome(script,"number");
    LatticeNeighbours.number[i]=readInInteger(script); 
    for(j=1;j<=LatticeNeighbours.number[i];j++) {
	  LatticeNeighbours.x[i][j]=readInInteger(script);
	  LatticeNeighbours.y[i][j]=readInInteger(script);
	  LatticeNeighbours.z[i][j]=readInInteger(script);
	  LatticeNeighbours.cell[i][j]=readInInteger(script);
    }
  }

  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=UnitCell.number;i++) {
      printf("Number of nearest-neighbours for site %d = %d : \n",
	    i, LatticeNeighbours.number[i]);
      for(j=1;j<=LatticeNeighbours.number[i];j++) {
        // ToDo: Neighbour = Bravais translation? wrong for ferriLC (JS06) 
	printf(" %d : Bravais translation (x y z site) = (% d % d % d % d)\n",
	       j, LatticeNeighbours.x[i][j],
                  LatticeNeighbours.y[i][j],
	          LatticeNeighbours.z[i][j],
                  LatticeNeighbours.cell[i][j]);
      }
      for(j=1;j<=LatticeNeighbours.number[i];j++) {
        // JS08: lc_j1j2j3 dim=1 but nearest-neighbours of uc in y direction
        //       cause error of recursiveClusteringFunction()
        if ((LatticeNeighbours.z[i][j] && Params.dim<3)
         || (LatticeNeighbours.y[i][j] && Params.dim<2)) {
          printf(" Error: dimension and lattice neighbours contradict\n");
          MPI_Abort(MPI_COMM_WORLD, 1);
        }
      }
    }
    printf("=======================================\n"
	   "The symmetry operations:\n");
  }
  moveAlongHome(script,"number");
  LatticeSymmetries.number=readInInteger(script); 

  if (LatticeSymmetries.number<1) {
    printf("input: we need at least the one symmetry operator (identity)\n"
           "  to get findDescriptorOfCluster running\nexit!\n");
    exit(1);
  }
  // just to make sure set to zero initially.

  for(i=1;i<=LatticeSymmetries.number;i++){
    // JS Jun08: change .x,.y,.z to LS.op[i].R[1..3][4]
    for(j=1;j<=3;j++)
      for(k=1;k<=4;k++) // ROT+TRANS
	LatticeSymmetries.R[i][j][k]=0.0;
  }
  for(i=1;i<=LatticeSymmetries.number;i++) {
    for(j=1;j<=3;j++)
      for(k=1;k<=4;k++)
	LatticeSymmetries.R[i][j][k]=readInFloat(script);
  }
  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=LatticeSymmetries.number;i++) {
      printf("Symmetry Operation = %d\nRotation matrix : Prim. Transl. \n",i);
      for(j=1;j<=3;j++){
	for(k=1;k<=3;k++)
	  printf("% 5.3f ",  LatticeSymmetries.R[i][j][k]);
	printf(" : % 5.3f\n",LatticeSymmetries.R[i][j][4]);
      }

      // ToDo: insert bravais-sym-computation here, BravaisSymmetries.B[][][]
      //  rotation in bravais coordinates + permutation of sites in unitcell
      //     for each site in unit cell per symmetry
      // (ROT,TRANS,UC) for each Sym*UC
      //  kagome?  mc_lf

    } 
  }

  // now determine the Hamiltonian. The `types' of term
  // first...
  
  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Hamiltonian:\n");
  
  // reset all terms
  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=0;i<MAX_UNIT_CELL;i++) {
      for(k=1+0;k<1+13;k++) {  // Jun08 Joerg: 13 types of terms 
	Hamiltonian[bonds][k].k_to_m_number[i]=0;
	Hamiltonian[bonds][k].m_to_k_number[i]=0;
        for(j=0;j<=MAX_NUM_NEIGHBOURS;j++) {
	  Hamiltonian[bonds][k].k_to_m[i][j][0]=0; // x
	  Hamiltonian[bonds][k].k_to_m[i][j][1]=0; // y
	  Hamiltonian[bonds][k].k_to_m[i][j][2]=0; // z
	  Hamiltonian[bonds][k].k_to_m[i][j][3]=0; // cell
	  Hamiltonian[bonds][k].m_to_k[i][j][0]=0; // x
	  Hamiltonian[bonds][k].m_to_k[i][j][1]=0; // y
	  Hamiltonian[bonds][k].m_to_k[i][j][2]=0; // z
	  Hamiltonian[bonds][k].m_to_k[i][j][3]=0; // cell
	  Hamiltonian[bonds][k].coefficient[i][j]=0.0;
	}
      }
    }
  }

  moveAlongHome(script,"directional");
  directional=readInInteger(script);
  //  1 for kagome_j1_j2, kagome_triangle, lc_mf, maple_bounce*,
  //        square_external_field, scrubo_spiral, triangle_external_field*

  if (myid==0)
    printf("Directional?=%d  (1=directional, 0=non-directional bonds)\n",
           directional);

  if ((!myid) && directional) // see ToDo below
     printf("\nWarning: Overwriting forward coefficient. "
            "If that causes error, report to JS!\n"); // Jun06

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=1;i<=UnitCell.number;i++) {
      for(k=1;k<=13;k++) {    /* 1..13 */

	moveAlongHome(script,"number");
	Hamiltonian[bonds][k].k_to_m_number[i]=readInInteger(script);
        if (Hamiltonian[bonds][k].k_to_m_number[i]>=MAX_NUM_NEIGHBOURS)
          ERR("numNeighbours>=MAX_NUM_NEIGHBOURS");
	
	if (k<=9) {  /* 1..9 are bonds */

          /* read in forward directions */
          for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
            Hamiltonian[bonds][k].k_to_m[i][j][0]=readInInteger(script);
            Hamiltonian[bonds][k].k_to_m[i][j][1]=readInInteger(script);
            Hamiltonian[bonds][k].k_to_m[i][j][2]=readInInteger(script);
            Hamiltonian[bonds][k].k_to_m[i][j][3]=readInInteger(script);
            Hamiltonian[bonds][k].coefficient[i][j]=readInFloat(script);
          }

	  if (directional) { /* two directions, read in backwards direction */

	    moveAlongHome(script,"number");
	    Hamiltonian[bonds][k].m_to_k_number[i]=readInInteger(script);
            if (Hamiltonian[bonds][k].m_to_k_number[i]>=MAX_NUM_NEIGHBOURS)
              ERR("numNeighbours>=MAX_NUM_NEIGHBOURS");

	    for(j=1;j<=Hamiltonian[bonds][k].m_to_k_number[i];j++) {
              float rr;
	      Hamiltonian[bonds][k].m_to_k[i][j][0]=readInInteger(script); // x
	      Hamiltonian[bonds][k].m_to_k[i][j][1]=readInInteger(script); // y
	      Hamiltonian[bonds][k].m_to_k[i][j][2]=readInInteger(script); // z
	      Hamiltonian[bonds][k].m_to_k[i][j][3]=readInInteger(script); // cell
              rr=readInFloat(script);
              // JS0812 kagome_j1_j2.ccm: forward-backward<=1e-9
	      if (fabs(rr-Hamiltonian[bonds][k].coefficient[i][j])>1e-8
                      &&  Hamiltonian[bonds][k].m_to_k_number[i]
                       <= Hamiltonian[bonds][k].k_to_m_number[i]
                      && !myid)
                printf("Warning: forward Coeff replaced by other backward"
                       " Coeff (bond=%d,type=%d,site=%d)\n",bonds,k,i);
	      Hamiltonian[bonds][k].coefficient[i][j]=rr;
              /* ToDo: overwiting same coefficient? */
	    }

	  } else {    
            /* one direction, copy forwards to backwards direction */

	    Hamiltonian[bonds][k].m_to_k_number[i]=
	      Hamiltonian[bonds][k].k_to_m_number[i];

	    for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
	      Hamiltonian[bonds][k].m_to_k[i][j][0]=Hamiltonian[bonds][k].k_to_m[i][j][0];
	      Hamiltonian[bonds][k].m_to_k[i][j][1]=Hamiltonian[bonds][k].k_to_m[i][j][1];
	      Hamiltonian[bonds][k].m_to_k[i][j][2]=Hamiltonian[bonds][k].k_to_m[i][j][2];
	      Hamiltonian[bonds][k].m_to_k[i][j][3]=Hamiltonian[bonds][k].k_to_m[i][j][3];	      
	    }
	  }
	} else {  /* 10..13 are single site terms */
	  Hamiltonian[bonds][k].coefficient[i][1]=readInFloat(script);
	}
      }
    }
  }

  /* set flag to evaluate if we have nonzero coefficient (gt 0.01) */
  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(k=1;k<=13;k++) {
      Hamiltonian[bonds][k].evaluate=0;
      for(i=1;i<=UnitCell.number;i++) {
	for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
//        until 0605 only coefficients bigger than fabs(0.01) are used
// Joerg Jun08 ToDo: what about m_to_k? a bug?
	  if (Hamiltonian[bonds][k].coefficient[i][j] != 0.0) {
  	      Hamiltonian[bonds][k].evaluate=1;
	      break;
	  }
	}
	if (Hamiltonian[bonds][k].evaluate) break;
      }
    }
  }

  // now print out this information:

    
  if (Control.output > 0 && myid ==0) {
    const char *tdescr[13+1]={"none",
      "s^z s^z",
      "s^z s^+", "s^z s^-", "s^+ s^z", "s^- s^z",
      "s^+ s^-", "s^- s^+", "s^+ s^+", "s^- s^-",
      "s^z", "s^-", "(s^z)^2", "s^+"};  // Oct05 Joerg: correct???
    for(bonds=1;bonds<=Params.bonds;bonds++) {
      printf("\nBond= %2d\n", bonds);
      for(k=1;k<=13;k++) {
        if (!Hamiltonian[bonds][k].evaluate) continue; // Oct05
        // space before next type of bonds
        if (k > 9) {  /* single site term */
          for(i=1;i<=UnitCell.number;i++) {
            if (Hamiltonian[bonds][k].coefficient[i][1])
            printf("Bond= %2d : Type= %2d : Site= %2d : Coeff= % .3f : %s\n",
                   bonds, k, i,
                   Hamiltonian[bonds][k].coefficient[i][1],
                   tdescr[k]);
          }
        } else {  /* k<=9 bonds */
          for(i=1;i<=UnitCell.number;i++) {
            for(j=1;j<=Hamiltonian[bonds][k].k_to_m_number[i];j++) {
              printf("Bond= %2d : Type= %2d : Site=%2d : k-m %d : xyzc= % d % d % d % d : Coeff= % .3f\n",
                     bonds, k, i, j,
                     Hamiltonian[bonds][k].k_to_m[i][j][0],
                     Hamiltonian[bonds][k].k_to_m[i][j][1],
                     Hamiltonian[bonds][k].k_to_m[i][j][2],
                     Hamiltonian[bonds][k].k_to_m[i][j][3],
                     Hamiltonian[bonds][k].coefficient[i][j]);
            }

            if (!directional) continue;
            for(j=1;j<=Hamiltonian[bonds][k].m_to_k_number[i];j++) {
              printf("Bond= %2d : Type= %2d : Site=%2d : m-k %d : xyzc= % d % d % d % d : Coeff= % .3f\n",
                   bonds, k, i, j,
                   Hamiltonian[bonds][k].m_to_k[i][j][0],
                   Hamiltonian[bonds][k].m_to_k[i][j][1],
                   Hamiltonian[bonds][k].m_to_k[i][j][2],
                   Hamiltonian[bonds][k].m_to_k[i][j][3],
                   Hamiltonian[bonds][k].coefficient[i][j]);
            }
          }
        }
      }
    }
  }

  // now input the information for the Hamiltonian proper for each term
  // in the similarity transformed version of H.
  // Jun08 Joerg: transformation removed for simplification

  fclose(script);

  // 5-5-08: Additional code to implement generalised expectation values 
  if (Control.greatExpectations > 0) {
    double dtmp;  // temporary for read in
  
    moveAlongHome(NULL,""); /* reset counter */

    // first: the number of expectation values
    moveAlongHome(escript,"number of expectations"); Params.nExpect=readInInteger(escript);
    if (Control.output > 0 && myid ==0)
      printf("\n\nNumber of Expectation Values = %d \n",Params.nExpect);
  
    if (Params.nExpect>MAX_BONDS) ERR("nExpects > MAX_BONDS");

    // Jun08 JS: zero init added, ToDo: replace by readBonds(File,expectation)
    for(expect=1;expect<=MAX_BONDS;expect++) 
      for(i=1;i<=UnitCell.number;i++) 
        for(k=1;k<=13;k++){
          expectation[expect][k].k_to_m_number[i]=0;
          expectation[expect][k].m_to_k_number[i]=0;
          for(j=0;j<=MAX_NUM_NEIGHBOURS;j++)
            expectation[expect][k].coefficient[i][j]=0.0;
        }


    for(expect=1;expect<=Params.nExpect;expect++) {
      for(i=1;i<=UnitCell.number;i++) {
        for(k=1;k<=13;k++) { /* 1..13 */

     	   moveAlongHome(escript,"number");
           expectation[expect][k].k_to_m_number[i]=readInInteger(escript);
           if (expectation[expect][k].k_to_m_number[i]>=MAX_NUM_NEIGHBOURS)
              ERR("numNeighbours>=MAX_NUM_NEIGHBOURS");
	
           if (k<=9) {  /* 1..9 are bonds */

             /* read in forward directions */
	     for(j=1;j<=expectation[expect][k].k_to_m_number[i];j++) {
               expectation[expect][k].k_to_m[i][j][0]=readInInteger(escript);
               expectation[expect][k].k_to_m[i][j][1]=readInInteger(escript);
               expectation[expect][k].k_to_m[i][j][2]=readInInteger(escript);
               expectation[expect][k].k_to_m[i][j][3]=readInInteger(escript);
               expectation[expect][k].coefficient[i][j]=readInFloat(escript);
             }
             /* one direction, copy forwards to backwards direction */
	     expectation[expect][k].m_to_k_number[i]=
	       expectation[expect][k].k_to_m_number[i];
             for(j=1;j<=expectation[expect][k].k_to_m_number[i];j++) {
               expectation[expect][k].m_to_k[i][j][0]=expectation[expect][k].k_to_m[i][j][0];
               expectation[expect][k].m_to_k[i][j][1]=expectation[expect][k].k_to_m[i][j][1];
               expectation[expect][k].m_to_k[i][j][2]=expectation[expect][k].k_to_m[i][j][2];
               expectation[expect][k].m_to_k[i][j][3]=expectation[expect][k].k_to_m[i][j][3];
             }
             if (!directional) continue; 
             /* two directions, read in backwards direction */

	     moveAlongHome(escript,"number");
	     expectation[expect][k].m_to_k_number[i]=readInInteger(escript);
             if (expectation[expect][k].m_to_k_number[i]>=MAX_NUM_NEIGHBOURS)
                ERR("numNeighbours>=MAX_NUM_NEIGHBOURS");

	     for(j=1;j<=expectation[expect][k].m_to_k_number[i];j++) {
	       expectation[expect][k].m_to_k[i][j][0]=readInInteger(escript); // x
	       expectation[expect][k].m_to_k[i][j][1]=readInInteger(escript); // y
	       expectation[expect][k].m_to_k[i][j][2]=readInInteger(escript); // z
	       expectation[expect][k].m_to_k[i][j][3]=readInInteger(escript); // cell
               dtmp=readInFloat(escript);
               if (j<=expectation[expect][k].k_to_m_number[i] && 
                      expectation[expect][k].coefficient[i][j]==dtmp)
                 printf("Warning: bond=%d term=%d j=%d forward coeff overwritten\n",expect,k,j);
	       expectation[expect][k].coefficient[i][j]=dtmp;
               /* ToDo: overwiting same coefficient? */
	     }

	  } else {  /* 10..13 are single site terms */
	      expectation[expect][k].coefficient[i][1]=readInFloat(escript);
     	}
      }
    }
  }

  /* set flag to evaluate if we have nonzero coefficient (gt 0.01) */
  for(expect=1;expect<=Params.nExpect;expect++) {
    for(k=1;k<=13;k++) {
      expectation[expect][k].evaluate=0;  // Jun08 JS: [bonds] => [expect]
      for(i=1;i<=UnitCell.number;i++) {
        // Jun08 Joerg: j not initialized, [i][j] to [i][1]
        if (k> 9 && expectation[expect][k].coefficient[i][1] != 0.0)
	            expectation[expect][k].evaluate=1;
        if (k<10 && expectation[expect][k].k_to_m_number[i] > 0)
	            expectation[expect][k].evaluate=1;
      }
    }
  }
	
 // now print out this information:

  if (Control.output > 0 && myid ==0) {
    const char *tdescr[14]={"none",
      "s^z s^z",
      "s^z s^+", "s^z s^-", "s^+ s^z", "s^- s^z",
      "s^+ s^-", "s^- s^+", "s^+ s^+", "s^- s^-",
      "s^z", "s^-", "(s^z)^2", "s^+"};  // Oct05 Joerg: correct???
    for(expect=1;expect<=Params.nExpect;expect++) {
      printf("\nExpect= %d\n", expect);
      for(k=1;k<14;k++) {
        if (!expectation[expect][k].evaluate) continue; // Oct05
        // space \n before next type of expect
        if (k > 9) {  // single site terms
          for(i=1;i<=UnitCell.number;i++) {
            if (expectation[expect][k].coefficient[i][1])
            printf("Expect= %d : Type= %2d : Site= %d : Coeff= % .3f : %s\n",
                   expect, k, i,
                   expectation[expect][k].coefficient[i][1], tdescr[k]);
          }
        } else { // k<10 bonds
          for(i=1;i<=UnitCell.number;i++) {
            for(j=1;j<=expectation[expect][k].k_to_m_number[i];j++) {
              printf("Expect= %d : Type= %2d : Site= %d %s %d : xyzc= % d % d % d % d : Coeff= % .3f : %s\n",
                     expect, k, i, ((directional)?"k-m":""), j,
                     expectation[expect][k].k_to_m[i][j][0], // x
                     expectation[expect][k].k_to_m[i][j][1], // y
                     expectation[expect][k].k_to_m[i][j][2], // z
                     expectation[expect][k].k_to_m[i][j][3], // site of unit cell
                     expectation[expect][k].coefficient[i][j], tdescr[k]);
            }
            if (!directional) continue;
            for(j=1;j<=expectation[expect][k].m_to_k_number[i];j++) {
              printf("Expect= %d : Type= %2d : Site= %d m-k %d : xyzc= % d % d % d % d : Coeff= % .3f : %s\n",
                     expect, k, i, j,
                     expectation[expect][k].m_to_k[i][j][0], // x
                     expectation[expect][k].m_to_k[i][j][1], // y
                     expectation[expect][k].m_to_k[i][j][2], // z
                     expectation[expect][k].m_to_k[i][j][3], // site of cell
                     expectation[expect][k].coefficient[i][j], tdescr[k]);
                     /* ToDo: do we need k_to_m and m_to_k coefficients? */
            }
          }
        }
      }
    }
  }

	
  }  // ControlGreatExpectations
  
  definePointsInBT();

  //  init x_array[],x_index[],cell_index[]
  //  defineGeneralisedPositions();  -> removed (not used, Joerg05)

  // not used at moment...

  sublatticeMagnetisation.number=UnitCell.number;
  for (i=1;i<=sublatticeMagnetisation.number;i++)
    sublatticeMagnetisation.cell[i]=1;
 
  // read in cluster configurations 
  
  if ((run_option&1) == 0) { // loading configs from file, else compute
    int nread;

    sprintf(configurations2, "%s.configs", Control.name);
    if (!myid) 
      printf("\nReading ConfigDescriptors(n*%d*(int=%d)) from file %s\n",
                      MAX_SPIN_LIST, (int)sizeof(int), configurations2);
    if (!myid) {  // JS Apr08: read only once and distribute per MPI
      fp1=fopen(configurations2,"r"); // $(Control.name).configs  
      if (!fp1) {
        printf("input: Cannot Open %s... Exiting!!\n",configurations2);
        exit(1);
      }
      // pre060611: fread(&NumberConfigurations,sizeof(int),1,fp1);    
      for(i=1;i<=MAX_NUM_CONFIGURATIONS;i++) {

        nread=fread(spinConfigurations[i].descriptor,
               ((MAX_SPIN_LIST)*sizeof(int)), 1, fp1);
        if (nread!=1) {
           if (feof(fp1)) { i--; break; }
           printf("%d: ferror %d\nExit!\n",myid,ferror(fp1));
           exit(1);
        }
        if (Control.output > 2 && myid==0 && i<10){
          printf("Cluster %d:\n",i);
          printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
        }
        if (feof(fp1)) break;
      }
      fclose(fp1);
    }
    NumberConfigurations = i;

    if (!myid) {
      printf("\nFundamental clusters = %d\n", NumberConfigurations);
    }
#if PARALLEL_FLAG>0
    int iret=0;
    if (Control.output>0 && !myid)
      printf(" calling MPI_Bcast(.configs, (int)(%d*%d)) ...\n",
             (int)sizeof(struct spinConfigs), i); 
    iret= MPI_Bcast(&NumberConfigurations, 1, MPI_INT, 0, MPI_COMM_WORLD);
    if (iret!=MPI_SUCCESS) MPI_Abort(MPI_COMM_WORLD, 1);
    i = NumberConfigurations;
    iret= MPI_Bcast(spinConfigurations+1,
               sizeof(struct spinConfigs)*i, MPI_BYTE, 0, MPI_COMM_WORLD);
    if (iret!=MPI_SUCCESS) MPI_Abort(MPI_COMM_WORLD, 1);
#endif

  }
}


// Joerg Jun08: return 0, if found, -1 if failed
int moveAlongHome(FILE *fp, const char *searchstring)
{
  char dummy, buff[4000];
  int  count, bp=0, iscomment=0;
  static int line=1;

  /* 
     purpose of this routine is to scan along the file until
     an `=' sign is seen 
     Joerg Sep05: ignore out commented lines (// ... and # ...)
     Joerg Jun08: add searchstring, for better debugging (new control parameter)
  */

  /* reset */  
  if (!fp) { line=1; return 0; }

  buff[0]=0;
  for(count=0;count>-1;count++) {
    if (feof(fp)) { count=-1; break; }
    fscanf(fp,"%c",&dummy);              // char by char
    if (dummy != '\n' && (dummy != '=' || iscomment)) {
      buff[bp]=dummy; bp++; buff[bp]=0; // add a char to the buffer
      if (bp>=4000) { count=-2; break; }
    }
    if (Control.output > 2 && myid == 0) {
      if (dummy == '\n' && buff[0])    printf("%% %s\n",buff);
      if (dummy == '='  && !iscomment) printf("%% %s= ...\n",buff);
    }
    if (bp>1 && buff[bp-1] == '/'
             && buff[bp-2] == '/') iscomment=1;
    if (dummy == '#')              iscomment=1;
    if (dummy == '\n') { iscomment=0; line++; bp=0; buff[bp]=0; }
    if (dummy == '=' && !iscomment) {
      /* empty searchstring OR hit */
      if ((!searchstring[0]) || strcmp(buff,searchstring)==0) 
        break;  // found, if searchstring=="" accept everything
      else {
        if (!myid)
          printf("ERROR: searched for \"%s=\" but found \"%s=\" (line=%d)\n",
                 searchstring, buff, line);
        ERR("bad file format");
      }
    }
  }
  if (count < 0) {
    ERR("Error in reading CCM script: can't find delimiter! Exiting!!");
  }
  return 0;
}

int readInInteger(FILE *fp) {
  int sum;
  fscanf(fp,"%d",&sum);
  return sum;
}
    
double readInFloat(FILE *fp) {
  double sum;
  fscanf(fp,"%lf",&sum);
  return sum;
}


extern int block_size;  // see initDivisors

// called by input() = initBTarray (BT = bounding triangle)
// initialize: lookUpTable[0..sizeLUT-1]  (old MAX_BT)
//                 used by [printOut,find]ClusterFromDescriptor
//             BTarray[z][y][x] // memory: m^3*int
// replaced by BTarray[z][y]+x  // memory: m^2*int
// replaced by BTarray[z*m+y]+x // malloc(m^(dim-1)*int) (ToDo?)
//                 used by findConfigurationToDescriptor
//                 0 <= x,y,z < Params.btLength
// ToDo: estimate size and malloc lookUpTable, simplify (Joerg)
//   sample: btLength=m=4 D=2
//   (y,x)=(0,0),(0,1),(0,2),(0,3), = 0,1,2,3,
//         (1,0),(1,1),(1,2),         4,5,6,
//         (2,0),(2,1),               7,8,
//         (3,0)                      9                 
//                                  
//    n*(n+1)/2 = 4*5/2 = 10 entries
//    1D:  n/1  2D: n*(n+1)/2, 3D: n*(n+1)*(n+2)/6
// rewritten by Joerg Oct05         
void definePointsInBT(void)
{

  int i, x,y,z, count,mx,my,mz;
  
  if (Params.btLength > MAX_LSUBM) {
    printf("Params.btLength=%d > MAX_LSUBM\n", Params.btLength);
    MPI_Abort(MPI_COMM_WORLD,1);
  }

  if (Params.btLength < Params.m) {
    printf("Params.btLength=%d < Params.m\n", Params.btLength);
    MPI_Abort(MPI_COMM_WORLD,1);
  }

  mx=my=mz=Params.btLength; // maximum x,y,z for 3D-case
  if (Params.dim < 3) mz=1; // max_z=0 for 2D (periodic 2D lattice)
  if (Params.dim < 2) my=1; // max_y=0 for 1D (periodic 1D lattice)
  if (Params.dim < 1) mx=1; // max_x=0 for 0D (finite cluster = UC)

  count=0; // site index, starting from 0 (Oct05)
  for (z=0;z<mz    ;z++) 
  for (y=0;y<my-z  ;y++) 
  for (x=0;x<mx-y-z;x++) { // (x,y,z) <=> idx, replacable by 2 functions ???
    count++;               
  }
  //  1D: n/1,  2D: n*(n+1)/(1*2),  3D: n*(n+1)*(n+2)/(1*2*3)
  { int count2=1;
    for (i=0;i<Params.dim;i++) count2 *= (Params.btLength+i);
    for (i=0;i<Params.dim;i++) count2 /= (i+1);
    if (count != count2) {   // can not happen! but we check for safety
      printf("ERROR definePointsInBT count!=count2\n");
      MPI_Abort(MPI_COMM_WORLD,1);
    }
  }
  sizeLUT = count;
  lookUpTable = (int (*)[3]) malloc(sizeLUT*sizeof(int)*3);
  if (!lookUpTable) {
      printf("ERROR definePointsInBT lookUpTable malloc(%d*int)\n", sizeLUT);
      MPI_Abort(MPI_COMM_WORLD,1);
  }

  BTarray = (int *) malloc( mz*my*sizeof(int) );
  if (!BTarray) {
      printf("ERROR definePointsInBT BTarray malloc(%d*int)\n", mz*my);
      MPI_Abort(MPI_COMM_WORLD,1);
  }

  count=0; // site index, starting from 0 (Oct05)
  for (z=0;z<mz    ;z++) 
  for (y=0;y<my-z  ;y++) 
  for (x=0;x<mx-y-z;x++) { // (x,y,z) <=> idx, replacable by 2 functions ???
    if (x==0) BTarray[z*Params.btLength+y]=count;
    lookUpTable[count][0]=x;
    lookUpTable[count][1]=y;
    lookUpTable[count][2]=z;
    count++;                    //  2D: n*(n+1)/2, 3D: n*(n+1)*(n+2)/6
  }
  // added by Joerg Sep2005  cubic 10 -> 10*12^2/4=1440/4=385=10*11/2*7
  // set this constants to reduce memory needs to minimum
  {
    int max_spin_list = (sizeLUT*UnitCell.number + block_size - 1)/block_size;
    if (Control.output > 0 && myid==0) {
      printf("\ndefinePointsInBT: btLength= %d sizeLUT= %d UC= %d BSize= %d\n",
          Params.btLength, sizeLUT, UnitCell.number, block_size);
      printf("\ndefinePointsInBT: MAX_SPIN_LIST= %d (Constants.h)\n",
          max_spin_list);
    }

    if (max_spin_list > MAX_SPIN_LIST) {
      printf("Error: Constant.h MAX_SPIN_LIST < %d\n", max_spin_list);
      MPI_Abort(MPI_COMM_WORLD, 1);
    }
  }
}  
